@@ -88,6 +88,7 @@ class FourSphereVolumeConductor(object):
             2.39290752e-10, 2.39290752e-10, 2.39290752e-10, 2.39290752e-10,
             2.39290752e-10, 2.39290752e-10]])
     """
+
     def __init__(self,
                  r_electrodes,
                  radii=[79000., 80000., 85000., 90000.],
@@ -171,10 +172,11 @@ class FourSphereVolumeConductor(object):
                  'Can be avoided by placing dipole further away from '
                  'brain surface.')
 
-        if any(r < self._rz for r in self.r):
-            raise RuntimeError('Electrode must be farther away from '
-                               'brain center than dipole: r > rz.'
-                               'r = %s, rz = %s', self.r, self._rz)
+        if any(r <= self._rz for r in self.r):
+            mssg = ('Electrode must be farther away from ' +
+                    'brain center than dipole: r > rz.' +
+                    f'r = {self.r}, rz = {self._rz}')
+            raise RuntimeError(mssg)
 
         # compute theta angle between rzloc and rxyz
         self._theta = self._calc_theta()
@@ -196,9 +198,9 @@ class FourSphereVolumeConductor(object):
         Returns
         -------
         potential: ndarray, dtype=float
-            Shape (n_contacts, n_timesteps) array containing the electric
-            potential at contact point(s) FourSphereVolumeConductor.r in units
-            of (mV) for all timesteps of current dipole moment p.
+            Shape ``(n_contacts, n_timesteps)`` array containing the electric
+            potential at contact point(s) ``FourSphereVolumeConductor.rxyz``
+            in units of (mV) for all timesteps of current dipole moment ``p``.
 
         """
 
@@ -206,17 +208,17 @@ class FourSphereVolumeConductor(object):
         n_contacts = self.r.shape[0]
         n_timesteps = p.shape[1]
 
-        if np.linalg.norm(p) != 0:
+        if np.linalg.norm(p) > 0:
             p_rad, p_tan = self._decompose_dipole(p)
         else:
             p_rad = np.zeros((3, n_timesteps))
             p_tan = np.zeros((3, n_timesteps))
-        if np.linalg.norm(p_rad) != 0.:
+        if np.linalg.norm(p_rad) > 0:
             pot_rad = self._calc_rad_potential(p_rad)
         else:
             pot_rad = np.zeros((n_contacts, n_timesteps))
 
-        if np.linalg.norm(p_tan) != 0.:
+        if np.linalg.norm(p_tan) > 0:
             pot_tan = self._calc_tan_potential(p_tan)
         else:
             pot_tan = np.zeros((n_contacts, n_timesteps))
@@ -227,8 +229,8 @@ class FourSphereVolumeConductor(object):
     def get_transformation_matrix(self, dipole_location):
         '''
         Get linear response matrix mapping current dipole moment in (nA µm)
-        located in location `rz` to extracellular potential in (mV)
-        at recording sites `FourSphereVolumeConductor.` (µm)
+        located in location ``rz`` to extracellular potential in (mV)
+        at recording sites ``FourSphereVolumeConductor.rxyz`` (µm)
 
         parameters
         ----------
@@ -370,57 +372,72 @@ class FourSphereVolumeConductor(object):
         """
         cos_theta = (self.rxyz @ self._rzloc) / (
             np.linalg.norm(self.rxyz, axis=1) * np.linalg.norm(self._rzloc))
-        theta = np.arccos(cos_theta)
-        return theta
+        # avoid RuntimeWarning: invalid value encountered in arccos
+        # and resulting NaNs
+        with np.errstate(invalid='ignore'):
+            theta = np.arccos(cos_theta)
+        return np.nan_to_num(theta)
 
     def _calc_phi(self, p_tan):
         """
-        Return azimuthal angle between x-axis and contact point locations(s)
+        Return azimuthal angle between u-axis and contact point locations(s)
+        for tangential dipole component ``p_tan``
+
+        w-axis is here defined in the direction of rzloc.
+        v-axis is here defined in the direction of ``p_tan``
+        (orthogonal to ``rzloc``).
+        u-axis is here defined as cross product between `
+        `p_tan`` and ``rzloc``.
 
         Parameters
         ----------
-        p_tan: ndarray, dtype=float
-            Shape (3, n_timesteps) array containing
+        ``p_tan``: ndarray, dtype=float
+            Shape ``(3, n_timesteps)`` array containing
             tangential component of current dipole moment in units of (nA*µm)
 
         Returns
         -------
-        phi: ndarray, dtype=float
-            Shape (n_contacts, n_timesteps) array containing azimuthal angle
-            in units of (radians) between x-axis vector(s) and projection of
-            contact point location vector(s) rxyz into xy-plane.
-            z-axis is defined in the direction of rzloc.
-            y-axis is defined in the direction of p_tan (orthogonal to rzloc).
-            x-axis is defined as cross product between p_tan and rzloc (x).
+        ``phi``: ndarray, dtype=float
+            Shape ``(n_contacts, n_timesteps)`` array containing azimuth angle
+            in units of (radians) between u-axis vector(s) and projection of
+            contact point location vector(s) ``rxyz`` into uv-plane.
         """
 
-        # project rxyz onto z-axis (rzloc)
-        proj_rxyz_rz = self.rxyz * self._z
-        # find projection of rxyz in xy-plane
-        rxy = self.rxyz - proj_rxyz_rz
-        # define x-axis
-        x = np.cross(p_tan.T, self._z)
+        # project rxyz onto w-axis (axis of rzloc)
+        proj_rxyz_rz = np.outer(self.rxyz @ self._rzloc / self._rz, self._z)
 
+        # find projection of rxyz in uv-plane
+        r_uv = self.rxyz - proj_rxyz_rz
+        # define u-axis
+        u = np.cross(p_tan.T, self._z)
+
+        # azimuth angle
         phi = np.zeros((len(self.rxyz), p_tan.shape[1]))
-        # create masks to avoid computing phi when phi is not defined
-        mask = np.ones(phi.shape, dtype=bool)
-        # phi is not defined when theta= 0,pi or |p_tan| = 0
-        mask[(self._theta == 0) | (self._theta == np.pi)] = np.zeros(
-            p_tan.shape[1])
-        mask[:, np.abs(np.linalg.norm(p_tan, axis=0)) == 0] = 0
 
-        cos_phi = np.zeros(phi.shape)
-        # compute cos_phi using mask to avoid zerodivision
-        cos_phi[mask] = (rxy @ x.T)[mask] \
-            / np.outer(np.linalg.norm(rxy, axis=1),
-                       np.linalg.norm(x, axis=1))[mask]
+        if np.linalg.norm(u) > 0:
+            # create masks to avoid computing phi when phi is not defined
+            mask = np.ones(phi.shape, dtype=bool)
+            # phi is not defined when theta= 0, pi or |p_tan| = 0
+            mask[(self._theta == 0) | (self._theta == np.pi)] = np.zeros(
+                p_tan.shape[1])
+            mask[:, np.abs(np.linalg.norm(p_tan, axis=0)) == 0] = False
 
-        # compute phi in [0, pi]
-        phi[mask] = np.arccos(cos_phi[mask])
+            cos_phi = np.zeros(phi.shape)
+            # compute cos_phi using mask to avoid zero division
+            cos_phi[mask] = (r_uv @ u.T)[mask] / \
+                np.outer(np.linalg.norm(r_uv, axis=1),
+                         np.linalg.norm(u, axis=1))[mask]
 
-        # nb: phi in [-pi, pi]. since p_tan defines direction of y-axis,
-        # phi < 0 when rxy*p_tan < 0
-        phi[(rxy @ p_tan) < 0] *= -1
+            # above linear operations may result in cos_phi outside [-1, +1]
+            cos_phi[cos_phi < -1.] = -1.
+            cos_phi[cos_phi > 1.] = 1.
+
+            # compute phi in [0, pi]
+            phi[mask] = np.nan_to_num(np.arccos(cos_phi[mask]))
+
+            # NB: phi in [-pi, pi]. since p_tan defines direction of y-axis,
+            # phi < 0 when r_uv * p_tan < 0
+            phi[(r_uv @ p_tan) < 0] *= -1
 
         return phi
 
@@ -885,9 +902,10 @@ class InfiniteVolumeConductor(object):
         return self.get_dipole_potential(np.eye(3), r)
 
 
-class MEG(object):
+class InfiniteHomogeneousVolCondMEG(object):
     """
-    Basic class for computing magnetic field from current dipole moment.
+    Basic class for computing magnetic field from current dipole moment in
+    an infinite homogeneous volume conductor model.
     For this purpose we use the Biot-Savart law derived from Maxwell's
     equations under the assumption of negligible magnetic induction
     effects [1]_:
@@ -919,6 +937,7 @@ class MEG(object):
 
     See also
     --------
+    SphericallySymmetricVolCondMEG
     FourSphereVolumeConductor
     InfiniteVolumeConductor
 
@@ -932,7 +951,7 @@ class MEG(object):
 
     >>> import LFPy, os, numpy as np, matplotlib.pyplot as plt
     >>> from lfpykit import CurrentDipoleMoment
-    >>> from lfpykit.eegmegcalc import MEG
+    >>> from lfpykit.eegmegcalc import InfiniteHomogeneousVolCondMEG as MEG
     >>> # create LFPy.Cell object
     >>> cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
     >>>                                          'ball_and_sticks.hoc'),
@@ -980,9 +999,10 @@ class MEG(object):
     AssertionError
         If dimensionality of sensor_locations is wrong
     """
+
     def __init__(self, sensor_locations, mu=4 * np.pi * 1E-7):
         """
-        Initialize class MEG
+        Initialize class InfiniteHomogeneousVolCondMEG
         """
         assert sensor_locations.ndim == 2, 'sensor_locations.ndim != 2'
         assert sensor_locations.shape[1] == 3, 'sensor_locations.shape[1] != 3'
@@ -1012,7 +1032,7 @@ class MEG(object):
     def calculate_H(self, current_dipole_moment, dipole_location):
         """
         Compute magnetic field H from single current-dipole moment localized
-        somewhere in space
+        in an infinite homogeneous volume conductor.
 
         Parameters
         ----------
@@ -1056,6 +1076,252 @@ class MEG(object):
 
         return H
 
+    def calculate_B(self, p, r_p):
+        """
+        Compute magnetic field B from single current dipole ``p`` localized
+        somewhere in space at ``r_p``.
+
+        This function returns the magnetic
+        field :math:`\\mathbf{B}=µ\\mathbf{H}`.
+
+        Parameters
+        ----------
+        p: ndarray, dtype=float
+            shape (3, n_timesteps) array with x,y,z-components of current-
+            dipole moment time series data in units of (nA µm)
+        r_p: ndarray, dtype=float
+            shape (3, ) array with x,y,z-location of dipole in units of (µm)
+
+        Returns
+        -------
+        ndarray, dtype=float
+            shape (n_locations x 3 x n_timesteps) array with x,y,z-components
+            of the magnetic field :math:`\\mathbf{B}` in units of (nA/µm)
+
+        """
+        return self.mu * self.calculate_H(p, r_p)
+
+
+class MEG(InfiniteHomogeneousVolCondMEG):
+    def __init__(self, sensor_locations, mu=4 * np.pi * 1E-7):
+        warn(
+            "class MEG is deprecated and will be removed. Use "
+            "InfiniteHomogeneousVolCondMEG or SphericallySymmetricVolCondMEG "
+            "instead",
+            DeprecationWarning, 2
+        )
+        super().__init__(sensor_locations, mu)
+
+
+class SphericallySymmetricVolCondMEG(object):
+    """Computes magnetic fields from current dipole in
+    spherically-symmetric volume conductor models.
+
+    This class facilitates calculations according to eq. (34) from [1]_
+    (see also [2]_) defined as:
+
+    .. math::
+
+        \\mathbf{H} = \\frac{1}{4 \\pi} \\frac{F \\mathbf{p} \\times
+        \\mathbf{r}_p - (\\mathbf{p} \\times \\mathbf{r}_p \\cdot
+        \\mathbf{r}) \\nabla F}{F^2},
+        \\text{ where}
+
+        F = a(ra + r^2 - \\mathbf{r}_p \\cdot \\mathbf{r}),
+
+        \\nabla F = (r^{-1}a^2 + a^{-1}\\mathbf{a}
+        \\cdot \\mathbf{r} + 2a + 2r)\\mathbf{r}
+        -(a + 2r + a^{-1}\\mathbf{a} \\cdot \\mathbf{r})\\mathbf{r}_p,
+
+        \\mathbf{a} = \\mathbf{r} - \\mathbf{r}_p,
+
+        a = |\\mathbf{a}|,
+
+        r = |\\mathbf{r}| .
+
+    Here,
+    :math:`\\mathbf{p}` is the current dipole moment,
+    :math:`\\mathbf{r}` the measurement location(s) and
+    :math:`\\mathbf{r}_p` the current dipole location.
+
+    Note that the magnetic field :math:`\\mathbf{H}` is related to the magnetic
+    field :math:`\\mathbf{B}` as
+
+    .. math:: \\mu_0 \\mathbf{H} = \\mathbf{B}-\\mathbf{M} ,
+
+    where :math:`\\mu_0` denotes the permeability of free space (very close to
+    permebility of biological tissues).
+    :math:`\\mathbf{M}` denotes material
+    magnetization (which is ignored).
+
+    Parameters
+    ----------
+    r: ndarray
+        sensor locations, shape ``(n, 3)`` where ``n`` denotes number of
+        locations, unit [µm]
+    mu: float
+        Permeability. Default is permeability of vacuum
+        (:math:`\\mu_0 = 4\\pi 10^{-7}` Tm/A)
+
+    See also
+    --------
+    InfiniteHomogeneousVolCondMEG
+
+    Examples
+    --------
+    Compute the magnetic field from a current dipole located
+
+    >>> import numpy as np
+    >>> from lfpykit.eegmegcalc import SphericallySymmetricVolCondMEG
+    >>> p = np.array([[0, 1, 0]]).T  # tangential current dipole (nAµm)
+    >>> r_p = np.array([0, 0, 90000])  # dipole location (µm)
+    >>> r = np.array([[0, 0, 92000]])  # measurement location (µm)
+    >>> m = SphericallySymmetricVolCondMEG(r=r)
+    >>> M = m.get_transformation_matrix(r_p=r_p)
+    >>> H = M @ p
+    >>> H  # magnetic field (nA/µm)
+    array([[[9.73094081e-09],
+            [0.00000000e+00],
+            [0.00000000e+00]]])
+
+    References
+    ----------
+    .. [1] Hämäläinen M., et al., Reviews of Modern Physics, Vol. 65, No. 2,
+        April 1993.
+    .. [2] Sarvas J., Phys.Med. Biol., 1987, Vol. 32, No 1, 11-22.
+
+    Raises
+    ------
+    AssertionError
+        If dimensionality of sensor locations ``r`` is wrong
+    """
+
+    def __init__(self, r, mu=4 * np.pi * 1E-7):
+        """
+        Initialize class SphericallySymmetricVolCondMEG
+        """
+        assert r.ndim == 2, 'r.ndim != 2'
+        assert r.shape[1] == 3, 'r.shape[1] != 3'
+
+        # set attributes
+        self.r = r
+        self.mu = mu
+
+    def _compute_F(self, r_p, r_i, a_n, r_n):
+        return a_n * (r_n * a_n + r_n**2 - r_p @ r_i.T)
+
+    def _compute_grad_F(self, r_p, r_i, a, a_n, r_n):
+        return ((a_n**2 / r_n + a @ r_i.T / a_n + 2 * a_n + 2 * r_n) * r_i
+                - (a_n + 2 * r_n + a @ r_i.T / a_n) * r_p)
+
+    def get_transformation_matrix(self, r_p):
+        """
+        Get linear response matrix mapping current dipole moment in (nA µm)
+        located in location ``r_p`` to magnetic field
+        :math:`\\mathbf{H}` in units of (nA/µm) at sensor locations ``r``
+
+        Parameters
+        ----------
+        r_p: ndarray, dtype=float
+            shape (3, ) array with x,y,z-location of dipole in units of (µm)
+
+        Returns
+        -------
+        response_matrix: ndarray
+            shape (n_sensors, 3, 3) ndarray
+
+        Raises
+        ------
+        AssertionError
+            If dipole location ``r_p`` has the wrong shape or if its radius
+            is greater than radius to any sensor location in ``<object>.r``
+
+        """
+        assert r_p.shape == (3, ), \
+            'r_p.shape != (3, )'
+
+        assert np.all(np.linalg.norm(self.r, axis=-1) > np.linalg.norm(r_p)), \
+            'the norm |r_p| must be less than that of any |r|'
+
+        # current dipole along x,y,z-axis
+        p = np.eye(3)
+        M = np.empty((self.r.shape[0], 3, 3))
+        for j, p_ in enumerate(p):
+            p_ = np.expand_dims(p_, 1)
+            for i, r_i in enumerate(self.r):
+                a = r_i - r_p
+                a_n = np.linalg.norm(a)
+                r_n = np.linalg.norm(r_i)
+                F = self._compute_F(r_p, r_i, a_n, r_n)
+                grad_F = self._compute_grad_F(r_p, r_i, a, a_n, r_n)
+                M[i, :, j] = ((F * np.cross(p_.T, r_p)
+                              - (np.cross(p_.T, r_p) @ r_i) * grad_F
+                               ).T / F**2 / 4 / np.pi).flatten()
+        return M
+
+    def calculate_H(self, p, r_p):
+        """
+        Compute magnetic field :math:`\\mathbf{H}` from single current dipole
+        ``p`` localized somewhere in space at ``r_p``
+
+        Parameters
+        ----------
+        p: ndarray, dtype=float
+            shape (3, n_timesteps) array with x,y,z-components of current-
+            dipole moment time series data in units of (nA µm)
+        r_p: ndarray, dtype=float
+            shape (3, ) array with x,y,z-location of dipole in units of (µm)
+
+        Returns
+        -------
+        ndarray, dtype=float
+            shape (n_locations x 3 x n_timesteps) array with x,y,z-components
+            of the magnetic field :math:`\\mathbf{H}` in units of (nA/µm)
+
+        Raises
+        ------
+        AssertionError
+            If dimensionality of current_dipole_moment ``p`` and/or
+            dipole_location ``r_p`` is wrong
+
+        """
+
+        assert p.ndim == 2, \
+            'p.ndim != 2'
+        assert p.shape[0] == 3, \
+            'p.shape[0] != 3'
+        assert r_p.shape == (3, ), \
+            'r_p.shape != (3, )'
+
+        M = self.get_transformation_matrix(r_p)
+
+        return M @ p
+
+    def calculate_B(self, p, r_p):
+        """
+        Compute magnetic field B from single current dipole ``p`` localized
+        somewhere in space at ``r_p``.
+
+        This function returns the magnetic
+        field :math:`\\mathbf{B}=µ\\mathbf{H}`.
+
+        Parameters
+        ----------
+        p: ndarray, dtype=float
+            shape (3, n_timesteps) array with x,y,z-components of current-
+            dipole moment time series data in units of (nA µm)
+        r_p: ndarray, dtype=float
+            shape (3, ) array with x,y,z-location of dipole in units of (µm)
+
+        Returns
+        -------
+        ndarray, dtype=float
+            shape (n_locations x 3 x n_timesteps) array with x,y,z-components
+            of the magnetic field :math:`\\mathbf{B}` in units of (nA/µm)
+
+        """
+        return self.mu * self.calculate_H(p, r_p)
+
 
 class NYHeadModel(object):
     """
@@ -1286,7 +1552,7 @@ class NYHeadModel(object):
         """
         if dipole_pos is None:
             dipole_pos_ = self.dipole_pos_dict['motorsensory_cortex']
-        elif type(dipole_pos) is str:
+        elif isinstance(dipole_pos, str):
             if dipole_pos not in self.dipole_pos_dict:
                 raise RuntimeError("When dipole_pos is string, location must"
                                    "be defined in self.dipole_pos_dict. "
@@ -1309,7 +1575,7 @@ class NYHeadModel(object):
         if loc_error > 2:
             raise RuntimeWarning("Large dipole location error! "
                                  "Given loc: {}; Closest vertex: {}".format(
-                                    dipole_pos_, self.dipole_pos))
+                                     dipole_pos_, self.dipole_pos))
 
         self.cortex_normal_vec = self.cortex_normals[:, self.vertex_idx]
 
